local addonName, addon = ...
local L = addon.L

HBD = LibStub("HereBeDragons-2.0")

-- Register events and call functions
addon.frame:SetScript("OnEvent", function(self, event, ...)
	if addon.debugging then print("LIME:", event, ...) end
	addon.frame[event](self, ...)
end)

local function init()
	if not addon.dataLoaded then addon.loadData() end
	if GuidelimeDataChar.mainFrameShowing then addon.showMainFrame() end
end

addon.frame:RegisterEvent('PLAYER_LOGIN')
function addon.frame:PLAYER_LOGIN()
	C_Timer.After(2, init)
end

addon.frame:RegisterEvent('PLAYER_LEVEL_UP')
function addon.frame:PLAYER_LEVEL_UP(level)
	C_Timer.After(0.1, function()
		addon.level = level
		addon.xpMax = UnitXPMax("player")
		if addon.debugging then print("LIME: You reached level " .. level .. ". Grats! new xp max is " .. addon.xpMax) end
		addon.updateSteps()
	end)
end

addon.frame:RegisterEvent('PLAYER_XP_UPDATE')
function addon.frame:PLAYER_XP_UPDATE(level)
	addon.xp = UnitXP("player")
	--if addon.debugging then print("LIME: xp is " .. addon.xp) end
	addon.updateSteps()
end

addon.frame:RegisterEvent('UPDATE_FACTION')
function addon.frame:UPDATE_FACTION(level)
	addon.updateMainFrame()
end

function addon.updateFromQuestLog()
	local questLog = {}
	local isCollapsed = {}
	local currentHeader
	if C_QuestLog.GetInfo ~= nil then
		local i = 1
		while (true) do	
			local info = C_QuestLog.GetInfo(i)
			if not info then break end
			if info.isHeader then
				isCollapsed[info.title] = info.isCollapsed
				currentHeader = info.title
			else
				questLog[info.questID] = {}
				questLog[info.questID].index = i
				--local completed = true
				--local objectives = C_QuestLog.GetQuestObjectives(info.questID)
				--for i, o in ipairs(objectives) do if not o.finished then completed = false end end
				questLog[info.questID].finished = C_QuestLog.IsComplete(info.questID)
				questLog[info.questID].failed = C_QuestLog.IsFailed(info.questID)
				questLog[info.questID].name = info.title
				questLog[info.questID].sort = currentHeader
			end
			i = i + 1
		end
	else
		for i = 1, GetNumQuestLogEntries() do
			local name, _, _, header, collapsed, completed, _, id = GetQuestLogTitle(i)
			if header then
				isCollapsed[name] = collapsed
				currentHeader = name
			else
				questLog[id] = {}
				questLog[id].index = i
				questLog[id].finished = (completed == 1)
				questLog[id].failed = (completed == -1)
				questLog[id].name = name
				questLog[id].sort = currentHeader	
			end
		end
	end
	
	local checkCompleted = false
	local questChanged = false
	local questFound = false
	local questItemsNeeded = {}
	for _, id in ipairs(addon.questIds) do
		local q = addon.quests[id]
		if questLog[id] ~= nil and not questLog[id].failed then
			local numObjectives = GetNumQuestLeaderBoards(questLog[id].index)
			if numObjectives == 0 then questLog[id].finished = true end
			if q.logIndex ~= nil then
				questFound = true
				if q.logIndex ~= questLog[id].index or q.finished ~= questLog[id].finished then
					questChanged = true
					q.logIndex = questLog[id].index
					q.finished = questLog[id].finished
					--if addon.debugging then print("LIME: changed log entry ".. id .. " finished", q.finished) end
				end
			else
				questFound = true
				questChanged = true
				q.logIndex = questLog[id].index
				q.finished = questLog[id].finished
				q.name = questLog[id].name
				q.sort = questLog[id].sort
				--if addon.debugging then print("LIME: new log entry ".. id .. " finished", q.finished) end
			end
			if q.objectives == nil or #q.objectives ~= numObjectives then q.objectives = {} end
			for k = 1, numObjectives do
				local desc, type, done = GetQuestLogLeaderBoard(k, addon.quests[id].logIndex)
				-- special treatment for item objectives: when the same item needs to be collected for different quests at the same time 
				-- all objectives should be done only when enough items for all quests have been collected
				if type == 'item' then
					local objectives = addon.getQuestObjectives(id)
					if objectives ~= nil and objectives[k] ~= nil and objectives[k].type == 'item' then
						local itemId = objectives[k].ids.item[1]
						local itemName, _, numNeeded = desc:match("([^%d]*)([%d]+)%s*/%s*([%d]+)")
						if numNeeded ~= nil then
							if questItemsNeeded[itemId] ~= nil then 
								local itemCount = GetItemCount(itemId) - questItemsNeeded[itemId]
								numNeeded = tonumber(numNeeded)
								if addon.debugging then print("LIME: item " .. itemId .. " " .. itemName .. " " .. itemCount .. "/" .. numNeeded) end
								if itemCount < numNeeded then
									done = false
									q.finished = false
									desc = itemName .. (itemCount >= 0 and itemCount or 0) .. "/" .. numNeeded
								end
							end
							questItemsNeeded[itemId] = (questItemsNeeded[itemId] or 0) + numNeeded
						else
							if addon.debugging then print("LIME: error parsing item objective text - " .. desc) end
						end
					end
				end
				if q.objectives[k] == nil or desc ~= q.objectives[k] or done ~= q.objectives[k].done then
					questChanged = true
					q.objectives[k] = {desc = desc, done = done, type = type}
				end					
			end
		elseif not q.completed then
			checkCompleted = true
			if q.logIndex ~= nil and q.logIndex ~= -1 and not isCollapsed[q.sort] then
				q.logIndex = nil
				--if addon.debugging then print("LIME: removed log entry ".. id) end
			end
		end
	end
	return checkCompleted, questChanged, questFound
end

local function doQuestUpdate()
	addon.xp = UnitXP("player")
	addon.xpMax = UnitXPMax("player")
	addon.x, addon.y, addon.instance = HBD:GetPlayerWorldPosition()
	
	if addon.quests ~= nil then 
		local checkCompleted, questChanged, questFound = addon.updateFromQuestLog()

		if addon.firstLogUpdate == nil then
			addon.updateMainFrame()
		else
			if not questChanged then
				if addon.contains(addon.currentGuide.steps, function(s) return not s.skip and not s.completed and s.active and s.xp ~= nil end) then 
					questChanged = true 
				end
			end
			
			if checkCompleted then
				if questFound then
					addon.updateStepsText()
				end
				C_Timer.After(0.1, function() 
					local completed = addon.GetQuestsCompleted()
					local questCompleted = false
					for id, q in pairs(addon.quests) do
						if completed[id] and not q.completed then
							questCompleted = true
							q.finished = true
							q.completed = true
						end
					end
					if questCompleted == true or GuidelimeDataChar.showCompletedSteps then
						addon.updateSteps()
					else
						-- quest was abandoned so redraw erverything since completed steps might have to be done again
						addon.updateMainFrame()
					end
				end)
			elseif questChanged then 
				addon.updateSteps() 
			elseif questFound then
				addon.updateStepsText()
			end
		end
	end
	addon.firstLogUpdate = true
end

addon.frame:RegisterEvent('QUEST_LOG_UPDATE')
function addon.frame:QUEST_LOG_UPDATE()
	doQuestUpdate()
end

function addon.isQuestAuto(option, id)
	if option == "All" then 
		return true 
	elseif addon.currentGuide == nil then
		return false
	elseif option == "Current" then 
		return addon.currentGuide.activeQuests ~= nil and addon.contains(addon.currentGuide.activeQuests, id)
	elseif option == "Guide" then
		return addon.scannedQuests ~= nil and addon.containsKey(addon.scannedQuests, id)
	else
		return false
	end
end


addon.frame:RegisterEvent('GOSSIP_SHOW')
function addon.frame:GOSSIP_SHOW()
	if (GuidelimeData.autoAcceptQuests or GuidelimeData.autoTurnInQuests) and not IsShiftKeyDown() then 
		if addon.debugging then print ("LIME: GOSSIP_SHOW", addon.GetGossipActiveQuests()) end
		if addon.debugging then print ("LIME: GOSSIP_SHOW", addon.GetGossipAvailableQuests()) end
		local q = { addon.GetGossipActiveQuests() }
		local selectActive = nil
		local selectAvailable = nil
		addon.openNpcAgain = false
		for i = 1, addon.GetNumGossipActiveQuests() do
			local name = q[(i-1) * 6 + 1]
			local complete = q[(i-1) * 6 + 4]
			if complete and addon.isQuestAuto(GuidelimeData.autoTurnInQuests, function(id) return name == addon.getQuestNameById(id) end) then
				if selectActive == nil then
					selectActive = i
				else
					addon.openNpcAgain = true
				end			
			end
		end
		q = { addon.GetGossipAvailableQuests() }
		for i = 1, addon.GetNumGossipAvailableQuests() do
			local name = q[(i-1) * 7 + 1]
			if addon.isQuestAuto(GuidelimeData.autoAcceptQuests, function(id) return name == addon.getQuestNameById(id) end) then
				if selectActive == nil and selectAvailable == nil then
					selectAvailable = i
				else
					addon.openNpcAgain = true
				end			
			end
		end

		if selectActive ~= nil then
			C_Timer.After(addon.AUTO_COMPLETE_DELAY, function() 
				if addon.debugging then print ("LIME: GOSSIP_SHOW selectActive", selectActive) end
				addon.SelectGossipActiveQuest(selectActive)
			end)
		elseif selectAvailable ~= nil then
			C_Timer.After(addon.AUTO_COMPLETE_DELAY, function() 
				if addon.debugging then print ("LIME: GOSSIP_SHOW selectAvailable", selectAvailable) end
				addon.SelectGossipAvailableQuest(selectAvailable)
			end)
		end
	end
	if GuidelimeData.autoSelectFlight and not IsShiftKeyDown() and addon.currentGuide ~= nil and addon.currentGuide.firstActiveIndex ~= nil and	addon.currentGuide.lastActiveIndex ~= nil then
		for i = addon.currentGuide.firstActiveIndex, addon.currentGuide.lastActiveIndex do
			local step = addon.currentGuide.steps[i]
			for _, element in ipairs(step.elements) do
				if not element.completed then
					if element.t == "FLY" or element.t == "GET_FLIGHT_POINT" then
						local gossip = {GetGossipOptions()}
						for i = 1, GetNumGossipOptions() do
							if gossip[i * 2] == "taxi" then
								addon.SelectGossipOption(i)
							end
						end
					end
				end
			end
		end
	end
end

addon.frame:RegisterEvent('QUEST_GREETING')
function addon.frame:QUEST_GREETING()
	if GuidelimeData.autoAcceptQuest and not IsShiftKeyDown() then 
		if addon.debugging then print ("LIME: QUEST_GREETING", GetNumActiveQuests()) end
		if addon.debugging then print ("LIME: QUEST_GREETING", GetNumAvailableQuests()) end
		local selectActive = nil
		local selectAvailable = nil
		addon.openNpcAgain = false
		for i = 1, GetNumActiveQuests() do
			local name = GetActiveTitle(i)
			if addon.isQuestAuto(GuidelimeData.autoAcceptQuests, function(id) return name == addon.getQuestNameById(id) end) then
				if selectActive == nil then
					selectActive = i
				else
					addon.openNpcAgain = true
				end			
			end
		end
		for i = 1, GetNumAvailableQuests() do
			local name = GetAvailableTitle(i)
			if addon.isQuestAuto(GuidelimeData.autoAcceptQuests, function(id) return name == addon.getQuestNameById(id) end) then
				if selectActive == nil and selectAvailable == nil then
					selectAvailable = i
				else
					addon.openNpcAgain = true
				end			
			end
		end

		if selectActive ~= nil then
			C_Timer.After(addon.AUTO_COMPLETE_DELAY, function() 
				if addon.debugging then print ("LIME: QUEST_GREETING selectActive", selectActive) end
				SelectActiveQuest(selectActive)
			end)
		elseif selectAvailable ~= nil then
			C_Timer.After(addon.AUTO_COMPLETE_DELAY, function() 
				if addon.debugging then print ("LIME: QUEST_GREETING selectAvailable", selectAvailable) end
				SelectAvailableQuest(selectAvailable)
			end)
		end
	end
end

addon.frame:RegisterEvent('QUEST_DETAIL')
function addon.frame:QUEST_DETAIL()
	local id = GetQuestID()
	if addon.debugging then print ("LIME: QUEST_DETAIL", id) end
	if not IsShiftKeyDown() and addon.isQuestAuto(GuidelimeData.autoAcceptQuests, id) then
		C_Timer.After(addon.AUTO_COMPLETE_DELAY, function()
			AcceptQuest()
			if addon.openNpcAgain then 
				--todo
			end
		end)
	end
end

addon.frame:RegisterEvent('QUEST_PROGRESS')
function addon.frame:QUEST_PROGRESS()
	local id = GetQuestID()
	if addon.debugging then print ("LIME: QUEST_PROGRESS", id) end
	if not IsShiftKeyDown() and IsQuestCompletable() and addon.isQuestAuto(GuidelimeData.autoTurnInQuests, id) then
		C_Timer.After(addon.AUTO_COMPLETE_DELAY, function() 
			CompleteQuest()
			if addon.openNpcAgain then 
				--todo
			end
		end)
	end
end

addon.frame:RegisterEvent('QUEST_COMPLETE')
function addon.frame:QUEST_COMPLETE()
	local id = GetQuestID()
	if not IsShiftKeyDown() and addon.isQuestAuto(GuidelimeData.autoTurnInQuests, id) then
		if addon.debugging then print ("LIME: QUEST_COMPLETE", id) end
		if (GetNumQuestChoices() <= 1) then
			C_Timer.After(addon.AUTO_COMPLETE_DELAY, function() 
		        GetQuestReward(1)
		    end)
		end
	end
end

addon.frame:RegisterEvent('CINEMATIC_START')
function addon.frame:CINEMATIC_START()
	if GuidelimeData.skipCutscenes then
		StopCinematic()
	end
end

addon.frame:RegisterEvent('TAXIMAP_OPENED')
function addon.frame:TAXIMAP_OPENED()
	if GuidelimeData.autoSelectFlight and not IsShiftKeyDown() and addon.currentGuide ~= nil and addon.currentGuide.firstActiveIndex ~= nil and	addon.currentGuide.lastActiveIndex ~= nil then
		for i = addon.currentGuide.firstActiveIndex, addon.currentGuide.lastActiveIndex do
			local step = addon.currentGuide.steps[i]
			for _, element in ipairs(step.elements) do
				if not element.completed then
					if element.flightmaster ~= nil then
						local master = addon.flightmasterDB[element.flightmaster]
						if addon.debugging then print("LIME: looking for", master.zone, master.place) end
						for j = 1, NumTaxiNodes() do
							--if addon.debugging then print("LIME: ", TaxiNodeName(j)) end
							if addon.isFlightmasterMatch(master, TaxiNodeName(j)) then
								if element.t == "FLY" and TaxiNodeGetType(j) == "REACHABLE" then
									if IsMounted() then Dismount() end -- dismount before using the flightpoint
									if addon.debugging then print ("LIME: Flying to " .. (master.place or master.zone)) end
									if _G["TaxiButton"..j] then TaxiNodeOnButtonEnter(_G["TaxiButton"..j]) end
									C_Timer.After(0.5, function()
										TakeTaxiNode(j)
									end)
								elseif element.t == "GET_FLIGHT_POINT" and TaxiNodeGetType(j) == "CURRENT" then
									addon.completeSemiAutomatic(element)
								end
								return
							end
						end
					end
				end
			end
		end
	end
end

addon.frame:RegisterEvent('PLAYER_CONTROL_LOST')
function addon.frame:PLAYER_CONTROL_LOST()
	C_Timer.After(1, function() 
		if UnitOnTaxi("player") then
			if addon.debugging then print ("LIME: UnitOnTaxi") end
			addon.completeSemiAutomaticByType("FLY")
		end
	end)
end

addon.frame:RegisterEvent('UI_INFO_MESSAGE')
function addon.frame:UI_INFO_MESSAGE(errorType, message)
	if message == ERR_NEWTAXIPATH then
		if addon.debugging then print ("LIME: ERR_NEWTAXIPATH") end
		addon.completeSemiAutomaticByType("GET_FLIGHT_POINT")
	end
end

addon.frame:RegisterEvent('HEARTHSTONE_BOUND')
function addon.frame:HEARTHSTONE_BOUND(errorType, message)
	addon.completeSemiAutomaticByType("SET_HEARTH")
end


addon.frame:RegisterEvent('UNIT_SPELLCAST_SUCCEEDED')
function addon.frame:UNIT_SPELLCAST_SUCCEEDED(unitTarget, castGUID, spellID)
	-- hearthstone was used (or Astral Recall)
	if spellID == 8690 or spellID == 556 then
		addon.completeSemiAutomaticByType("HEARTH")
	end
end

addon.requestItemInfo = {}
function addon.GetItemInfo(id)
	itemName, itemLink, itemQuality = GetItemInfo(id)
	if not itemName then
		if addon.requestItemInfo[id] == false or addon.requestItemInfo[id] == 50 then
			addon.requestItemInfo[id] = false
		else
			addon.requestItemInfo[id] = (addon.requestItemInfo[id] or 0) + 1
		end
	end
	return itemName, itemLink, itemQuality
end
addon.frame:RegisterEvent('GET_ITEM_INFO_RECEIVED')
function addon.frame:GET_ITEM_INFO_RECEIVED(itemId,success)
	if addon.requestItemInfo[itemId] and success then
		addon.requestItemInfo[itemId] = nil
		addon.updateStepsText()
		addon.updateUseItemButtons()
	end
end

addon.frame:RegisterEvent('BAG_UPDATE')
function addon.frame:BAG_UPDATE()
	local guide = GuidelimeDataChar and addon.guides[GuidelimeDataChar.currentGuide]
	if guide and guide.itemUpdateIndices and #guide.itemUpdateIndices > 0 then
		addon.updateSteps(guide.itemUpdateIndices)
	else
		addon.updateUseItemButtons()
	end
end

addon.frame:RegisterEvent('PLAYER_FARSIGHT_FOCUS_CHANGED')
function addon.frame:PLAYER_FARSIGHT_FOCUS_CHANGED()
	-- This is a hack to work around an issue with HereBeDragons
	-- see https://www.curseforge.com/wow/addons/herebedragons/issues/31
	-- In situations where player is remote controlling coordinates reported by hbd stay on the player character position and minimap icons do not move
	-- Therefore arrow and minimap will be disabled
	-- TODO: Are there other similar "remote-controlled" quest vehicles?
	local spellIds = {
		6196,   -- Far Sight, 
		321297, -- Eyes of the Beast
		6197,   -- Eagle Exe
		51852,  -- Eye of Acherus (in DK intro quest)
	}
	local i = 1
	while (true) do
		local name, _, _, _, _, _, _, _, _, spellId = UnitAura("player", i, "HELPFUL")
		if not name then break end
		--if addon.debugging then print ("LIME: player has buff", name, spellId) end
		if addon.contains(spellIds, spellId) then
			if not addon.hideMinimapIconsAndArrowWhileBuffed then
				if addon.debugging then print ("LIME: deactivating arrow and minimap due to", name) end
				addon.hideMinimapIconsAndArrowWhileBuffed = true
				addon.updateStepsMapIcons()
			end
			return
		end
		i = i + 1
	end
	if addon.hideMinimapIconsAndArrowWhileBuffed then
		if addon.debugging then print ("LIME: reactivating arrow and minimap") end
		addon.hideMinimapIconsAndArrowWhileBuffed = false
		addon.updateStepsMapIcons()
	end
end

addon.frame:RegisterEvent('PLAYER_REGEN_ENABLED')
function addon.frame:PLAYER_REGEN_ENABLED()
	if addon.updateAfterCombat then
		addon.updateAfterCombat = false
		addon.updateUseItemButtons()
	end
end

addon.frame:RegisterEvent('UPDATE_BINDINGS')
function addon.frame:UPDATE_BINDINGS()
	addon.updateUseItemButtons()
end

addon.frame:RegisterEvent('BAG_UPDATE_COOLDOWN')
function addon.frame:BAG_UPDATE_COOLDOWN()
	if not addon.mainFrame or not addon.mainFrame.useButtons then return end
	for _, button in ipairs(addon.mainFrame.useButtons) do
		if button:IsShown() then
			button:Update()
		end
	end
end
